"""
Web component
"""

import logging
import os
from datetime import datetime
from functools import wraps
from typing import Any, Callable

import yaml
from flask import Blueprint, Response, render_template, request
from flask_wtf import FlaskForm  # type: ignore[import]
from wtforms import HiddenField  # type: ignore[import]
from wtforms import DateTimeLocalField, IntegerField, StringField
from wtforms.validators import DataRequired, Length  # type: ignore[import]

from slinky import Slinky

slinky_webapp = Blueprint('webapp', __name__, template_folder='templates')

with open('config.yaml', encoding='utf-8-sig') as conffile:
    cfg = yaml.safe_load(conffile)


class DelForm(FlaskForm):  # type: ignore[misc]
    """
    Delete form definition
    """

    delete = HiddenField('delete')


class AddForm(FlaskForm):  # type: ignore[misc]
    """
    Add form definition
    """

    url = StringField(
        'URL',
        validators=[DataRequired(), Length(1, 2048)],
        render_kw={
            'size': 64,
            'maxlength': 2048,
            'placeholder': 'https://www.example.com',
        },
    )
    fixed_views = IntegerField(
        'Fixed number of views',
        validators=[DataRequired()],
        render_kw={
            'size': 3,
            'value': -1,
        },
    )
    length = IntegerField(
        'Shortcode length',
        validators=[DataRequired()],
        render_kw={
            'size': 3,
            'value': 4,
        },
    )
    expiry = DateTimeLocalField(
        'Expiry',
        format='%Y-%m-%dT%H:%M',
        render_kw={
            'size': 8,
            'maxlength': 10,
        },
    )


def protect(func: Callable[..., Response]) -> Callable[..., Response]:
    """
    Decorator that will protect the admin interface

    Args:
        func (Callable): Wrapped function

    Returns:
        Callable: Function wrapper
    """

    @wraps(func)
    def check_ip(*args: Any, **kwargs: Any) -> Response:
        if (
            os.environ.get('FLASK_ENV', '') != 'development'
            and request.headers['X-Forwarded-For'] not in cfg['allowed_ips']
        ):
            print(f'Protected URL access attempt from {request.remote_addr}')
            return Response('Not found', 404)
        return func(*args, **kwargs)

    return check_ip


@slinky_webapp.route('/<path:path>', strict_slashes=False)
def try_path_as_shortcode(path: str) -> Response:
    """
    Try the initial path as a shortcode, redirect if found

    Returns:
        Response: redirect if found, otherwise 404
    """
    should_redirect = True
    slinky = Slinky(cfg['db'])
    shortcode = slinky.get_by_shortcode(path)
    if shortcode.url:
        if shortcode.fixed_views == 0:
            logging.warning('Shortcode out of views')
            should_redirect = False
        elif shortcode.fixed_views > 0:
            slinky.remove_view(shortcode.id)
        if datetime.fromisoformat(shortcode.expiry) < datetime.now():
            logging.warning('Shortcode expired')
            should_redirect = False

        if should_redirect:
            return Response(
                'Redirecting...', status=302, headers={'location': shortcode.url}
            )

    return Response('Not found', 404)


@slinky_webapp.route('/_/add', methods=['GET', 'POST'])
@protect
def add() -> Response:
    """
    Create and add a new shorturl

    Returns:
        Response: HTTP response
    """
    shortcode = ''
    url = ''

    form = AddForm(meta={'csrf': False})

    if form.is_submitted():
        url = form.url.data.strip()
        length = form.length.data
        fixed_views = form.fixed_views.data
        expiry = form.expiry.data or datetime.max

    if url:
        slinky = Slinky(cfg['db'])
        for attempts in range(50):
            try:
                shortcode = slinky.add(url, length, fixed_views, expiry)
                break
            except ValueError:
                logging.warning(
                    'Shortcode already exists.  Retrying (%d/50).', attempts
                )
        else:
            return Response('Could not create a unique shortcode', 500)

    return Response(render_template('add.html', form=form, shortcode=shortcode), 200)


@slinky_webapp.route('/_/list', methods=['GET', 'POST'])
@protect
def lister() -> Response:
    """
    List the shortcodes, URLs, etc.

    Returns:
        Response: HTTP response
    """
    form = DelForm(meta={'csrf': False})
    slinky = Slinky(cfg['db'])

    if form.is_submitted():
        slinky.delete_by_shortcode(form.delete.data.strip())

    return Response(
        render_template('list.html', form=form, shortcodes=slinky.get_all()),
        200,
    )


@slinky_webapp.route('/_/edit/<int:id>', methods=['GET', 'POST'])
@protect
def edit(id: int) -> Response:  # pylint: disable=invalid-name,redefined-builtin
    """
    Edit the shortcode.

    Returns:
        Response: HTTP response
    """
    form = DelForm(meta={'csrf': False})
    slinky = Slinky(cfg['db'])

    logging.debug('Editing: %d', id)

    if form.is_submitted():
        slinky.delete_by_shortcode(form.delete.data.strip())

    return Response(
        render_template('edit.html', form=form, shortcodes=slinky.get_all()),
        200,
    )
